@@ -5,10 +5,11 @@ Base class for finding modules.
 import dis
 import imp
 import marshal
-import new
 import opcode
 import os
+import pkgutil
 import sys
+import types
 import zipfile
 
 import cx_Freeze.hooks
@@ -18,6 +19,8 @@ INPLACE_ADD = opcode.opmap["INPLACE_ADD"]
 LOAD_CONST = opcode.opmap["LOAD_CONST"]
 IMPORT_NAME = opcode.opmap["IMPORT_NAME"]
 IMPORT_FROM = opcode.opmap["IMPORT_FROM"]
+IMPORT_STAR = opcode.opmap["IMPORT_STAR"]
+STORE_FAST = opcode.opmap["STORE_FAST"]
 STORE_NAME = opcode.opmap["STORE_NAME"]
 STORE_GLOBAL = opcode.opmap["STORE_GLOBAL"]
 STORE_OPS = (STORE_NAME, STORE_GLOBAL)
@@ -26,10 +29,18 @@ __all__ = [ "Module", "ModuleFinder" ]
 
 class ModuleFinder(object):
 
-    def __init__(self, includeFiles, excludes, path, replacePaths):
+    def __init__(self, includeFiles = None, excludes = [], path = None,
+            replacePaths = None, copyDependentFiles = True, bootstrap = False,
+            compress = True):
         self.includeFiles = includeFiles
+        if includeFiles is None:
+            self.includeFiles = []
         self.excludes = dict.fromkeys(excludes)
         self.replacePaths = replacePaths
+        if replacePaths is None:
+            self.replacePaths = []
+        self.copyDependentFiles = copyDependentFiles
+        self.compress = compress
         self.path = path or sys.path
         self.modules = []
         self.aliases = {}
@@ -39,6 +50,35 @@ class ModuleFinder(object):
         self._zipFileEntries = {}
         self._zipFiles = {}
         cx_Freeze.hooks.initialize(self)
+        initialExcludedModules = self.excludes.copy()
+        self._AddBaseModules()
+        if not bootstrap:
+            self._ClearBaseModuleCode(initialExcludedModules)
+
+    def _AddBaseModules(self):
+        """Add the base modules to the finder. These are the modules that
+           Python imports itself during initialization and, if not found,
+           can result in behavior that differs from running from source;
+           also include modules used within the bootstrap code"""
+        self.ExcludeModule("cStringIO")
+        self.ExcludeModule("doctest")
+        self.ExcludeModule("getopt")
+        self.ExcludeModule("logging")
+        self.ExcludeModule("re")
+        self.ExcludeModule("subprocess")
+        self.IncludeModule("traceback")
+        self.IncludeModule("warnings")
+        self.IncludePackage("encodings")
+        if sys.version_info[0] >= 3:
+            self.IncludeModule("io")
+        if sys.version_info[:2] >= (3, 3):
+            self.AddAlias("_frozen_importlib", "importlib._bootstrap")
+            self.IncludeModule("_frozen_importlib")
+        if self.copyDependentFiles:
+            self.IncludeModule("imp")
+            self.IncludeModule("os")
+            self.IncludeModule("sys")
+            self.IncludeModule("zlib")
 
     def _AddModule(self, name):
         """Add a module to the list of modules but if one is already found,
@@ -52,6 +92,22 @@ class ModuleFinder(object):
                 del self._badModules[name]
         return module
 
+    def _ClearBaseModuleCode(self, initialExcludedModules):
+        """Clear the code for all of the base modules. This is done when not in
+           bootstrap mode so that the base modules are not included in the
+           zip file."""
+        for name in self.excludes:
+            if name in initialExcludedModules:
+                continue
+            del self._modules[name]
+        self.excludes = initialExcludedModules
+        for module in self._modules.values():
+            if module is None:
+                continue
+            if module.code is not None:
+                module.code = None
+                module.file = None
+
     def _DetermineParent(self, caller):
         """Determine the parent to use when searching packages."""
         if caller is not None:
@@ -62,25 +118,28 @@ class ModuleFinder(object):
     def _EnsureFromList(self, caller, packageModule, fromList,
             deferredImports):
         """Ensure that the from list is satisfied. This is only necessary for
-           package modules. If the caller is the package itself, actually
-           attempt to import right then since it must be a submodule; otherwise
-           defer until after all global names are defined in order to avoid
-           spurious complaints about missing modules."""
-        if caller is not packageModule:
-            deferredImports.append((packageModule, fromList))
+           package modules. If the package module has not been completely
+           imported yet, defer the import until it has been completely imported
+           in order to avoid spurious errors about missing modules."""
+        if packageModule.inImport and caller is not packageModule:
+            deferredImports.append((caller, packageModule, fromList))
         else:
-            if fromList == ("*",):
-                fromList = packageModule.allNames
             for name in fromList:
                 if name in packageModule.globalNames:
                     continue
                 subModuleName = "%s.%s" % (packageModule.name, name)
                 self._ImportModule(subModuleName, deferredImports, caller)
 
-    def _FindModule(self, name, path):
+    def _FindModule(self, name, path, namespace):
         try:
             return imp.find_module(name, path)
         except ImportError:
+            if namespace and name in sys.modules:
+                module = sys.modules[name]
+                info = ("", "", imp.PKG_DIRECTORY)
+                return None, module.__path__[0], info
+            if path is None:
+                path = []
             for location in path:
                 if name in self._zipFileEntries:
                     break
@@ -116,40 +175,54 @@ class ModuleFinder(object):
 
     def _ImportAllSubModules(self, module, deferredImports, recursive = True):
         """Import all sub modules to the given package."""
-        suffixes = dict.fromkeys([s[0] for s in imp.get_suffixes()])
-        for dir in module.path:
+        suffixes = [s[0] for s in imp.get_suffixes()]
+        for path in module.path:
             try:
-                fileNames = os.listdir(dir)
+                fileNames = os.listdir(path)
             except os.error:
                 continue
             for fileName in fileNames:
-                name, ext = os.path.splitext(fileName)
-                if ext not in suffixes:
-                    continue
-                if name == "__init__":
-                    continue
+                fullName = os.path.join(path, fileName)
+                if os.path.isdir(fullName):
+                    initFile = os.path.join(fullName, "__init__.py")
+                    if not os.path.exists(initFile):
+                        continue
+                    name = fileName
+                else:
+                    # We need to run through these in order to correctly pick
+                    # up PEP 3149 library names (e.g. .cpython-32mu.so).
+                    for suffix in suffixes:
+                        if fileName.endswith(suffix):
+                            name = fileName[:-len(suffix)]
+                            break
+                    else:
+                        continue
+                    if name == "__init__":
+                        continue
                 subModuleName = "%s.%s" % (module.name, name)
                 subModule, returnError = \
                         self._InternalImportModule(subModuleName,
                                 deferredImports)
                 if returnError and subModule is None:
-                    raise ImportError, "No module named %s" % subModuleName
+                    raise ImportError("No module named %r" % subModuleName)
                 module.globalNames[name] = None
                 if subModule.path and recursive:
                     self._ImportAllSubModules(subModule, deferredImports,
                             recursive)
 
-    def _ImportDeferredImports(self, deferredImports):
+    def _ImportDeferredImports(self, deferredImports, skipInImport = False):
         """Import any sub modules that were deferred, if applicable."""
         while deferredImports:
             newDeferredImports = []
-            for packageModule, subModuleNames in deferredImports:
-                self._EnsureFromList(packageModule, packageModule,
-                        subModuleNames, newDeferredImports)
+            for caller, packageModule, subModuleNames in deferredImports:
+                if packageModule.inImport and skipInImport:
+                    continue
+                self._EnsureFromList(caller, packageModule, subModuleNames,
+                        newDeferredImports)
             deferredImports = newDeferredImports
 
     def _ImportModule(self, name, deferredImports, caller = None,
-            relativeImportIndex = 0):
+            relativeImportIndex = 0, namespace = False):
         """Attempt to find the named module and return it or None if no module
            by that name could be found."""
 
@@ -157,7 +230,7 @@ class ModuleFinder(object):
         # the name given is the only name that will be searched
         if relativeImportIndex == 0:
             module, returnError = self._InternalImportModule(name,
-                    deferredImports)
+                    deferredImports, namespace = namespace)
 
         # old style relative import (only possibility in Python 2.4 and prior)
         # the name given is tried in all parents until a match is found and if
@@ -167,13 +240,13 @@ class ModuleFinder(object):
             while parent is not None:
                 fullName = "%s.%s" % (parent.name, name)
                 module, returnError = self._InternalImportModule(fullName,
-                        deferredImports)
+                        deferredImports, namespace = namespace)
                 if module is not None:
                     parent.globalNames[name] = None
                     return module
                 parent = self._GetParentByName(parent.name)
             module, returnError = self._InternalImportModule(name,
-                    deferredImports)
+                    deferredImports, namespace = namespace)
 
         # new style relative import (available in Python 2.5 and up)
         # the index indicates how many levels to traverse and only that level
@@ -193,12 +266,12 @@ class ModuleFinder(object):
             else:
                 name = "%s.%s" % (parent.name, name)
                 module, returnError = self._InternalImportModule(name,
-                        deferredImports)
+                        deferredImports, namespace = namespace)
 
         # if module not found, track that fact
         if module is None:
             if caller is None:
-                raise ImportError, "No module named %s" % name
+                raise ImportError("No module named %r" % name)
             self._RunHook("missing", name, caller)
             if returnError and name not in caller.ignoreNames:
                 callers = self._badModules.setdefault(name, {})
@@ -206,7 +279,7 @@ class ModuleFinder(object):
 
         return module
 
-    def _InternalImportModule(self, name, deferredImports):
+    def _InternalImportModule(self, name, deferredImports, namespace = False):
         """Internal method used for importing a module which assumes that the
            name given is an absolute name. None is returned if the module
            cannot be found."""
@@ -217,6 +290,7 @@ class ModuleFinder(object):
         if name in self._builtinModules:
             module = self._AddModule(name)
             self._RunHook("load", module.name, module)
+            module.inImport = False
             return module, False
         pos = name.rfind(".")
         if pos < 0:
@@ -226,9 +300,12 @@ class ModuleFinder(object):
         else:
             parentName = name[:pos]
             parentModule, returnError = \
-                    self._InternalImportModule(parentName, deferredImports)
+                    self._InternalImportModule(parentName, deferredImports,
+                            namespace = namespace)
             if parentModule is None:
                 return None, returnError
+            if namespace:
+                parentModule.ExtendPath()
             path = parentModule.path
             searchName = name[pos + 1:]
         if name in self.aliases:
@@ -238,32 +315,41 @@ class ModuleFinder(object):
             self._modules[name] = module
             return module, returnError
         try:
-            fp, path, info = self._FindModule(searchName, path)
+            fp, path, info = self._FindModule(searchName, path, namespace)
+            module = self._LoadModule(name, fp, path, info, deferredImports,
+                    parentModule, namespace)
         except ImportError:
             self._modules[name] = None
             return None, True
-        module = self._LoadModule(name, fp, path, info, deferredImports,
-                parentModule)
         return module, False
 
     def _LoadModule(self, name, fp, path, info, deferredImports,
-            parent = None):
+            parent = None, namespace = False):
         """Load the module, given the information acquired by the finder."""
         suffix, mode, type = info
         if type == imp.PKG_DIRECTORY:
-            return self._LoadPackage(name, path, parent, deferredImports)
+            return self._LoadPackage(name, path, parent, deferredImports,
+                    namespace)
         module = self._AddModule(name)
         module.file = path
         module.parent = parent
         if type == imp.PY_SOURCE:
-            module.code = compile(fp.read() + "\n", path, "exec")
+            if sys.version_info[0] >= 3:
+                import tokenize
+                fp = open(path, "rb")
+                encoding, lines = tokenize.detect_encoding(fp.readline)
+                fp = open(path, "U", encoding = encoding)
+            codeString = fp.read()
+            if codeString and codeString[-1] != "\n":
+                codeString = codeString + "\n"
+            module.code = compile(codeString, path, "exec")
         elif type == imp.PY_COMPILED:
             if isinstance(fp, str):
                 magic = fp[:4]
             else:
                 magic = fp.read(4)
             if magic != imp.get_magic():
-                raise ImportError, "Bad magic number in %s" % path
+                raise ImportError("Bad magic number in %s" % path)
             if isinstance(fp, str):
                 module.code = marshal.loads(fp[8:])
                 module.inZipFile = True
@@ -279,14 +365,21 @@ class ModuleFinder(object):
                 module.code = self._ReplacePathsInCode(topLevelModule,
                         module.code)
             self._ScanCode(module.code, module, deferredImports)
+        module.inImport = False
         return module
 
-    def _LoadPackage(self, name, path, parent, deferredImports):
+    def _LoadPackage(self, name, path, parent, deferredImports, namespace):
         """Load the package, given its name and path."""
         module = self._AddModule(name)
         module.path = [path]
-        fp, path, info = imp.find_module("__init__", module.path)
-        self._LoadModule(name, fp, path, info, deferredImports, parent)
+        try:
+            fp, path, info = imp.find_module("__init__", module.path)
+            self._LoadModule(name, fp, path, info, deferredImports, parent)
+        except ImportError:
+            if not namespace:
+                raise
+            fileName = os.path.join(path, "__init__.py")
+            module.code = compile("", fileName, "exec")
         return module
 
     def _ReplacePathsInCode(self, topLevelModule, co):
@@ -308,10 +401,17 @@ class ModuleFinder(object):
         for i, value in enumerate(constants):
             if isinstance(value, type(co)):
                 constants[i] = self._ReplacePathsInCode(topLevelModule, value)
-        return new.code(co.co_argcount, co.co_nlocals, co.co_stacksize,
-                co.co_flags, co.co_code, tuple(constants), co.co_names,
-                co.co_varnames, newFileName, co.co_name, co.co_firstlineno,
-                co.co_lnotab, co.co_freevars, co.co_cellvars)
+        if sys.version_info[0] < 3:
+            return types.CodeType(co.co_argcount, co.co_nlocals,
+                    co.co_stacksize, co.co_flags, co.co_code, tuple(constants),
+                    co.co_names, co.co_varnames, newFileName, co.co_name,
+                    co.co_firstlineno, co.co_lnotab, co.co_freevars,
+                    co.co_cellvars)
+        return types.CodeType(co.co_argcount, co.co_kwonlyargcount,
+                co.co_nlocals, co.co_stacksize, co.co_flags, co.co_code,
+                tuple(constants), co.co_names, co.co_varnames, newFileName,
+                co.co_name, co.co_firstlineno, co.co_lnotab, co.co_freevars,
+                co.co_cellvars)
 
     def _RunHook(self, hookName, moduleName, *args):
         """Run hook for the given module if one is present."""
@@ -320,7 +420,7 @@ class ModuleFinder(object):
         if method is not None:
             method(self, *args)
 
-    def _ScanCode(self, co, module, deferredImports):
+    def _ScanCode(self, co, module, deferredImports, topLevel = True):
         """Scan code, looking for imported modules and keeping track of the
            constants that have been created in order to better tell which
            modules are truly missing."""
@@ -328,11 +428,18 @@ class ModuleFinder(object):
         arguments = []
         code = co.co_code
         numOps = len(code)
+        is3 = sys.version_info[0] >= 3
         while opIndex < numOps:
-            op = ord(code[opIndex])
+            if is3:
+                op = code[opIndex]
+            else:
+                op = ord(code[opIndex])
             opIndex += 1
             if op >= dis.HAVE_ARGUMENT:
-                opArg = ord(code[opIndex]) + ord(code[opIndex + 1]) * 256
+                if is3:
+                    opArg = code[opIndex] + code[opIndex + 1] * 256
+                else:
+                    opArg = ord(code[opIndex]) + ord(code[opIndex + 1]) * 256
                 opIndex += 2
             if op == LOAD_CONST:
                 arguments.append(co.co_consts[opArg])
@@ -344,25 +451,45 @@ class ModuleFinder(object):
                     relativeImportIndex = -1
                     fromList, = arguments
                 if name not in module.excludeNames:
-                    subModule = self._ImportModule(name, deferredImports,
+                    importedModule = self._ImportModule(name, deferredImports,
                             module, relativeImportIndex)
-                    if subModule is not None:
-                        module.globalNames.update(subModule.globalNames)
-                        if fromList and subModule.path is not None:
-                            self._EnsureFromList(module, subModule, fromList,
-                                    deferredImports)
-            elif op == IMPORT_FROM:
+                    if importedModule is not None:
+                        if fromList and fromList != ("*",) \
+                                and importedModule.path is not None:
+                            self._EnsureFromList(module, importedModule,
+                                    fromList, deferredImports)
+            elif op == IMPORT_FROM and topLevel:
+                if is3:
+                    op = code[opIndex]
+                    opArg = code[opIndex + 1] + code[opIndex + 2] * 256
+                else:
+                    op = ord(code[opIndex])
+                    opArg = ord(code[opIndex + 1]) + \
+                            ord(code[opIndex + 2]) * 256
                 opIndex += 3
+                if op == STORE_FAST:
+                    name = co.co_varnames[opArg]
+                else:
+                    name = co.co_names[opArg]
+                storeName = True
+                if deferredImports:
+                    deferredCaller, deferredPackage, deferredFromList = \
+                            deferredImports[-1]
+                    storeName = deferredCaller is not module
+                if storeName:
+                    module.globalNames[name] = None
+            elif op == IMPORT_STAR and topLevel and importedModule is not None:
+                module.globalNames.update(importedModule.globalNames)
+                arguments = []
             elif op not in (BUILD_LIST, INPLACE_ADD):
-                if op in STORE_OPS:
+                if topLevel and op in STORE_OPS:
                     name = co.co_names[opArg]
-                    if name == "__all__":
-                        module.allNames.extend(arguments)
                     module.globalNames[name] = None
                 arguments = []
         for constant in co.co_consts:
             if isinstance(constant, type(co)):
-                self._ScanCode(constant, module, deferredImports)
+                self._ScanCode(constant, module, deferredImports,
+                        topLevel = False)
 
     def AddAlias(self, name, aliasFor):
         """Add an alias for a particular module; when an attempt is made to
@@ -382,20 +509,22 @@ class ModuleFinder(object):
             moduleName = name
         info = (ext, "r", imp.PY_SOURCE)
         deferredImports = []
-        module = self._LoadModule(moduleName, file(path, "U"), path, info,
+        module = self._LoadModule(moduleName, open(path, "U"), path, info,
                 deferredImports)
         self._ImportDeferredImports(deferredImports)
         return module
 
     def IncludeFiles(self, sourcePath, targetPath):
         """Include the files in the given directory in the target build."""
-        self.includeFiles.append((sourcePath, targetPath))
+        if self.copyDependentFiles:
+            self.includeFiles.append((sourcePath, targetPath))
 
-    def IncludeModule(self, name):
+    def IncludeModule(self, name, namespace = False):
         """Include the named module in the frozen executable."""
         deferredImports = []
-        module = self._ImportModule(name, deferredImports)
-        self._ImportDeferredImports(deferredImports)
+        module = self._ImportModule(name, deferredImports,
+                namespace = namespace)
+        self._ImportDeferredImports(deferredImports, skipInImport = True)
         return module
 
     def IncludePackage(self, name):
@@ -405,19 +534,52 @@ class ModuleFinder(object):
         module = self._ImportModule(name, deferredImports)
         if module.path:
             self._ImportAllSubModules(module, deferredImports)
-        self._ImportDeferredImports(deferredImports)
+        self._ImportDeferredImports(deferredImports, skipInImport = True)
         return module
 
     def ReportMissingModules(self):
         if self._badModules:
-            print "Missing modules:"
-            names = self._badModules.keys()
+            sys.stdout.write("Missing modules:\n")
+            names = list(self._badModules.keys())
             names.sort()
             for name in names:
-                callers = self._badModules[name].keys()
+                callers = list(self._badModules[name].keys())
                 callers.sort()
-                print "?", name, "imported from", ", ".join(callers)
-            print
+                sys.stdout.write("? %s imported from %s\n" % \
+                        (name, ", ".join(callers)))
+            sys.stdout.write("\n")
+
+    def WriteSourceFile(self, fileName):
+        dirName = os.path.dirname(fileName)
+        if not os.path.isdir(dirName):
+            os.makedirs(dirName)
+        outfp = open(fileName, "w")
+        names = list(self._modules.keys())
+        names.sort()
+        modulesWritten = []
+        for name in names:
+            module = self._modules[name]
+            if module is None or module.code is None:
+                continue
+            mangledName = "__".join(name.split("."))
+            sys.stdout.write("adding base module named %s\n" % name)
+            code = marshal.dumps(module.code)
+            size = len(code)
+            if module.path:
+                size = -size
+            modulesWritten.append((name, mangledName, size))
+            outfp.write("unsigned char M_%s[] = {" % mangledName)
+            for i in range(0, len(code), 16):
+                outfp.write("\n\t")
+                for op in code[i:i + 16]:
+                    if not isinstance(op, int):
+                        op = ord(op)
+                    outfp.write("%d," % op)
+            outfp.write("\n};\n\n");
+        outfp.write("static struct _frozen gFrozenModules[] = {\n")
+        for name, mangledName, size in modulesWritten:
+            outfp.write('    {"%s", M_%s, %d},\n' % (name, mangledName, size))
+        outfp.write("    {0, 0, 0}\n};\n")
 
 
 class Module(object):
@@ -431,8 +593,8 @@ class Module(object):
         self.globalNames = {}
         self.excludeNames = {}
         self.ignoreNames = {}
-        self.allNames = []
         self.inZipFile = False
+        self.inImport = True
 
     def __repr__(self):
         parts = ["name=%s" % repr(self.name)]
@@ -448,6 +610,11 @@ class Module(object):
     def ExcludeName(self, name):
         self.excludeNames[name] = None
 
+    def ExtendPath(self):
+        self.path = pkgutil.extend_path(self.path, self.name)
+        if self.parent is not None:
+            self.parent.ExtendPath()
+
     def IgnoreName(self, name):
         self.ignoreNames[name] = None
 
