@@ -6,10 +6,8 @@
 # the Win32 extensions behave as expected.
 #------------------------------------------------------------------------------
 
-import encodings
 import os
 import sys
-import warnings
 import zipimport
 
 sys.frozen = True
@@ -28,7 +26,8 @@ else:
 code = importer.get_code(moduleName)
 exec code in m.__dict__
 
-if sys.version_info[:2] >= (2, 5):
+versionInfo = sys.version_info[:3]
+if versionInfo >= (2, 5, 0) and versionInfo <= (2, 6, 4):
     module = sys.modules.get("threading")
     if module is not None:
         module._shutdown()
